/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include "dcbd.h"
#include "dcb_driver_if_types.h"
#include "dcb_driver_interface.h"
#include "dcb_protocol.h"
#include "dcb_types.h"
#include "drv_cfg.h"
#include "lldp/ports.h"

static int get_gnl_id(void);

static __u16 fid = 0;
static int nl_sd = 0;

static int init_socket(void)
{
	int sd;
	int rcv_size;
	struct sockaddr_nl snl;

	sd = socket(AF_NETLINK, SOCK_RAW, NETLINK_GENERIC);
	if (sd < 0)
		return sd;

	if (setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &rcv_size, sizeof(int)) < 0) {
		close(sd);
		return -EIO;
	}

	memset((void *)&snl, 0, sizeof(struct sockaddr_nl));
	snl.nl_family = AF_NETLINK;
	snl.nl_pid = getpid();

	if (bind(sd, (struct sockaddr *)&snl, sizeof(struct sockaddr_nl)) < 0) {
		close(sd);
		return -EIO;
	}

	return sd;
}

static struct nlmsghdr *start_msg(__u16 family_id, __u8 genl_cmd)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;

	/* nlh needs to be free'd by send_msg() */
	nlh = (struct nlmsghdr *)malloc(MAX_MSG_SIZE);
	if (NULL == nlh)
		return NULL;
	memset((void *)nlh, 0, MAX_MSG_SIZE);

	nlh->nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
	nlh->nlmsg_type = family_id;
	nlh->nlmsg_flags = NLM_F_REQUEST;
	nlh->nlmsg_seq = 0;
	nlh->nlmsg_pid = getpid();
	glh = NLMSG_DATA(nlh);
	glh->cmd = genl_cmd;
	glh->version = 1;

	return nlh;
}

static struct nlattr *add_nla(struct nlmsghdr *nlh, __u16 nla_type,
                              void *attr, __u16 nla_len)
{
	struct nlattr *nla;

	nla = (struct nlattr *)((char *)nlh + nlh->nlmsg_len);
	nla->nla_type = nla_type;
	nla->nla_len = nla_len + NLA_HDRLEN;
	if (attr)
		memcpy(NLA_DATA(nla), attr, nla_len);
	nlh->nlmsg_len += NLMSG_ALIGN(nla->nla_len);

	return nla;
}

/* free's nlh which was allocated by start_msg */
static int send_msg(struct nlmsghdr *nlh)
{
	struct sockaddr_nl nladdr;
	void *buf = (void *)nlh;
	int r, len = nlh->nlmsg_len;

	if (nlh == NULL)
		return 1;

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	
	do {
		r = sendto(nl_sd, buf, len, 0, (struct sockaddr *)&nladdr,
			sizeof(nladdr));
		TRACE1("GNETLINK: send_msg: sendto = ", r);

	} while (r < 0 && errno == EINTR);

	free(nlh);

	if (r < 0)
		return 1;
	else
		return 0;
}

static struct nlmsghdr *get_msg(void)
{
	struct nlmsghdr *nlh;
	int len;

	/* nlh needs to be free'd by caller */
	nlh = (struct nlmsghdr *)malloc(MAX_MSG_SIZE);
	if (NULL == nlh)
		return NULL;
	memset(nlh, 0, MAX_MSG_SIZE);

	len = recv(nl_sd, (void *)nlh, MAX_MSG_SIZE, 0);
	TRACE1("nlmsg_type is ", nlh->nlmsg_type)
	if ((nlh->nlmsg_type == NLMSG_ERROR) || (len < 0) ||
		!(NLMSG_OK(nlh, len))) {
		free(nlh);
		return NULL;
	}

	return nlh;
}

static int recv_msg(int cmd, int attr)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla;
	int rval;

	nlh = get_msg();

	if (NULL == nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	if ((glh->cmd != cmd) || (nla->nla_type != attr)) {
		TRACE("Bad netlink message attribute.");
		return -EIO;
	}

	rval = *(__u8 *)NLA_DATA(nla);
	free(nlh);
	return rval;
}

static int get_gnl_id(void)
{
	struct nlmsghdr *nlh;
	struct nlattr *nla;
	int rval = 0;

	nlh = start_msg(GENL_ID_CTRL, CTRL_CMD_GETFAMILY);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, CTRL_ATTR_FAMILY_NAME, IXGBE_NLTYPE_DCB_NAME,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	if (send_msg(nlh)) {
		TRACE("Error in sending get_gnl_id request.\n")
		return -EIO;
	}

	nlh = get_msg();
	if (NULL == nlh) {
		TRACE("Error in geting get_gnl_id reply.")
		return -EIO;
	}

	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	/* genetlink puts family name first, so go to next attribute.
	 * This is where a full genetlink library would be useful.
	*/
	nla = (struct nlattr *)((char *)nla + NLA_ALIGN(nla->nla_len));

	if (nla->nla_type != CTRL_ATTR_FAMILY_ID)
		rval = -EIO;
	else
		fid = *(__u16 *)NLA_DATA(nla);

	free(nlh);

	return rval;
}

static int get_state(char *ifname, __u8 *state)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla;
	int rval = 0;

	nlh = start_msg(fid, DCB_C_GSTATE);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(ifname) + 1);

	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (NULL == nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_GSTATE) {
		rval = -EIO;
	} else if (nla->nla_type != DCB_A_STATE) {
		/* Do we really want to code up an attribute parser?? */
		TRACE("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		rval = -EIO;
	} else {
		*state = *(__u8 *)NLA_DATA(nla);
	}

	free(nlh);

	return rval;
}


static int set_state(char *ifname, __u8 state)
{
	struct nlmsghdr *nlh;

	nlh = start_msg(fid, DCB_C_SSTATE);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(ifname) + 1);
	add_nla(nlh, DCB_A_STATE, (void *)&state, sizeof(__u8));

	if (send_msg(nlh))
		return -EIO;

	return(recv_msg(DCB_C_SSTATE, DCB_A_STATE));
}


/* returns: 0 on success
 *          1 on failure
*/
static int set_pfc_cfg(char *ifname, __u8 *pfc)
{
	struct nlmsghdr *nlh;
	struct nlattr *nla_parent, *nla_child;
	int i;

	printf("set_pfc_cfg: %s\n", ifname);
	nlh = start_msg(fid, DCB_C_PFC_SCFG);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(ifname) + 1);
	nla_parent = add_nla(nlh, DCB_A_PFC_CFG, NULL, 0);
	for (i = PFC_A_UP_0; i < PFC_A_UP_MAX; i++) {
		nla_child = add_nla(nlh, i, (void *)pfc, sizeof(__u8));
		nla_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
		pfc++;
	}

	if (send_msg(nlh))
		return -EIO;

	return(recv_msg(DCB_C_PFC_SCFG, DCB_A_PFC_CFG));
}

/* returns: 0 on success
 *          1 on failure
*/
static int set_pg(char *ifname, struct tc_config *tc, __u8 *bwg, int cmd)
{
	struct nlmsghdr *nlh;
	struct nlattr *class_parent, *param_parent, *nla_child;
	__u8 *p = (__u8 *)tc;
	__u8 *b = (__u8 *)bwg;
	int i, j;

	nlh = start_msg(fid, cmd);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(ifname) + 1);
	class_parent = add_nla(nlh, DCB_A_PG_CFG, NULL, 0);
	for (i = PG_A_TC_0; i < PG_A_TC_MAX; i++) {
		param_parent = add_nla(nlh, i, NULL, 0);
		for (j = TC_A_PARAM_STRICT_PRIO; j < TC_A_PARAM_MAX; j++) {
			nla_child = add_nla(nlh, j, (void *)p, sizeof(__u8));
			param_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
			p++;
		}
		class_parent->nla_len += NLA_ALIGN(param_parent->nla_len);
	}
	for (i = PG_A_BWG_0; i < PG_A_BWG_MAX; i++) {
		nla_child = add_nla(nlh, i, (void *)b, sizeof(__u8));
		class_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
		b++;
	}

	if (send_msg(nlh))
		return -EIO;

	return(recv_msg(cmd, DCB_A_PG_CFG));
}


/* returns 0: on error getting family id
 *         1: if successful
*/
int init_gn(void)
{
	int err = 0;

	if ((nl_sd = init_socket()) < 0) {
		fprintf(stderr, "Error creating Netlink socket\n");
		return 0;
	}

	err = get_gnl_id();
	if (err) {
		return 0;
	}
	TRACE1("family id \n", fid)

	return 1;
 }

int get_hw_state(char *device_name, int *dcb_state)
{
	int err = 0;
	__u8 state = 0;

	err = get_gnl_id();
	if (err)
		return err;

	err = get_state(device_name, &state);
	if (err)
		fprintf(stderr, "Error getting DCB state: %s\n", device_name);

	*dcb_state = state;

	return err;
}
	
int set_hw_state(char *device_name, int dcb_state)
{
	int err = 0;

	err = get_gnl_id();
	if (err)
		return err;

	return set_state(device_name, (__u8)dcb_state);
}
	
/* returns:  0 on success
 *           1 on failure
*/
int set_hw_pg(char *device_name, pgroup_attribs *pg_data, boolean_t oper_mode)
{
	int i;
	int rval = 0;
	struct tc_config tc[MAX_TRAFFIC_CLASSES];
	__u8 bwg[MAX_BANDWIDTH_GROUPS];
	pg_attribs        pg_df_store;
	pgroup_attribs    pg_df_data, *pg_temp;


	/*        code assumes 1:1 up2tc - will need to change when that
	 *        assumption is invalid
	 */
	if (!oper_mode) { /* oper mode is false */
		get_pg(DEF_CFG_STORE, &pg_df_store);
		memcpy(&pg_df_data.rx, &pg_df_store.rx, sizeof(pg_df_data.rx));
		memcpy(&pg_df_data.tx, &pg_df_store.tx, sizeof(pg_df_data.tx));
		pg_temp = &pg_df_data;
	} else {
		pg_temp = pg_data;
	}

	for (i = 0; i < MAX_TRAFFIC_CLASSES; i++) {
		tc[i].prio_type = pg_temp->tx.up[i].strict_priority;
		tc[i].bwg_id = pg_temp->tx.up[i].bandwidth_group_mapping;
		tc[i].bwg_percent =
			pg_temp->tx.up[i].percent_of_bandwidth_group_cap;
		tc[i].up_to_tc_bitmap =
			(u8)(1<<pg_temp->tx.up[i].traffic_class_mapping);
		bwg[i] = pg_temp->tx.bwg_percent_of_link[i];
	}

	rval = set_pg(device_name, &tc[0], &bwg[0], DCB_C_PGTX_SCFG);

	for (i = 0; i < MAX_TRAFFIC_CLASSES; i++) {
		tc[i].prio_type = pg_temp->rx.up[i].strict_priority;
		tc[i].bwg_id = pg_temp->rx.up[i].bandwidth_group_mapping;
		tc[i].bwg_percent =
			pg_temp->rx.up[i].percent_of_bandwidth_group_cap;
		tc[i].up_to_tc_bitmap =
			(u8)(1<<pg_temp->rx.up[i].traffic_class_mapping);
		bwg[i] = pg_temp->rx.bwg_percent_of_link[i];
	}

	rval |= set_pg(device_name, &tc[0], &bwg[0], DCB_C_PGRX_SCFG);

	return rval;
}

/* returns:  0 on success
 *           1 on failure
*/
int set_hw_pfc(char *device_name, dcb_pfc_list_type pfc_data, boolean_t oper_mode)
{
	int i;
	__u8 pfc[MAX_TRAFFIC_CLASSES];
	pfc_attribs             pfc_df_store;
	pfc_type                *pfc_temp;

	if (!oper_mode) /* oper mode is false */
	{
		get_pfc(DEF_CFG_STORE, &pfc_df_store);
		pfc_temp = pfc_df_store.admin;
	} else {
		pfc_temp = pfc_data;
	}

	for (i = 0; i < MAX_TRAFFIC_CLASSES; i++) {
		if (pfc_temp[i])
			pfc[i] = pfc_enabled;
		else
			pfc[i] = pfc_disabled;
	}

	return set_pfc_cfg(device_name, &pfc[0]);
}

int set_hw_all(char *device_name)
{
	struct nlmsghdr *nlh;
	int status = 1; /* status is always true */
	int retval = -EIO;

	nlh = start_msg(fid, DCB_C_SET_ALL);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)device_name,
	        strlen(device_name) + 1);
	add_nla(nlh, DCB_A_SET_ALL, (void *)&status, sizeof(__u8));

	if (send_msg(nlh))
		return -EIO;

	retval = recv_msg(DCB_C_SET_ALL, DCB_A_SET_ALL);

	/* driver will respond with 
	 * 0 = hw config changes made
	 * 1 = no hw config changes were necessary
	*/
	if (retval > 0)
		retval = 0;
	else
		set_port_hw_resetting(device_name, 1);

	return 0;
}

int check_gn_id(void)
{
	return fid;
}

int get_perm_hwaddr(const char *ifname, u8 *buf)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla_parent, *nla_child;
	int rval = 0;
	int i;

	nlh = start_msg(fid, DCB_C_GPERM_HWADDR);
	if (NULL == nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(ifname) + 1);
	nla_parent = add_nla(nlh, DCB_A_PERM_HWADDR, NULL, 0);
	nla_child = add_nla(nlh, PERM_HW_A_ALL, NULL, 0);
	nla_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (NULL == nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla_parent = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_GPERM_HWADDR) {
		printf("Hmm, this is not the message we were expecting.\n");
		rval = -EIO;
		goto get_error;
	}
	if (nla_parent->nla_type != DCB_A_PERM_HWADDR) {
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		rval = -EIO;
		goto get_error;
	}
	nla_child = NLA_DATA(nla_parent);
	nla_parent = (struct nlattr *)((char *)nla_parent +
	                               NLA_ALIGN(nla_parent->nla_len));

	for (i = 0; nla_parent > nla_child; i++) {
		*(buf+i) = *(__u8 *)NLA_DATA(nla_child);
		nla_child = (struct nlattr *)((char *)nla_child +
		                              NLA_ALIGN(nla_child->nla_len));
	}
	if (nla_parent != nla_child)
		printf("nla pointers are off\n");

get_error:
	free(nlh);
	return 0;
}

boolean_t check_port_dcb_mode(char *device_name)
{
	int dcb_state = 0, err;

	err = get_gnl_id();

	if (err) {
		fprintf(stderr, "Error getting family id\n");
		dcb_state = 0;
	} else {
		if (get_hw_state(device_name, &dcb_state))
			dcb_state = 0;
	}

	if (dcb_state) {
		TRACE2("config.c: %s dcb mode is ON.\n", device_name);
		return TRUE;
	} else {
		TRACE2("config.c: %s dcb mode is OFF. \n", device_name);
		return FALSE;
	}
}
